package dev.fitko.fitconnect.cli.keygen;

import com.nimbusds.jose.JWEAlgorithm;
import com.nimbusds.jose.JWSAlgorithm;
import com.nimbusds.jose.jwk.JWK;
import com.nimbusds.jose.jwk.KeyOperation;
import com.nimbusds.jose.jwk.KeyType;
import dev.fitko.fitconnect.core.crypto.JWECryptoService;
import dev.fitko.fitconnect.jwkvalidator.JWKValidator;
import dev.fitko.fitconnect.jwkvalidator.exceptions.LogLevel;
import org.hamcrest.CoreMatchers;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;

import java.nio.charset.StandardCharsets;
import java.util.Map;

import static org.hamcrest.MatcherAssert.assertThat;
import static org.hamcrest.Matchers.contains;
import static org.hamcrest.Matchers.hasSize;
import static org.hamcrest.Matchers.is;
import static org.hamcrest.Matchers.notNullValue;
import static org.hamcrest.Matchers.nullValue;
import static org.junit.jupiter.api.Assertions.assertDoesNotThrow;

class JWKGeneratorTest {

    private JWKGenerator underTest;

    @BeforeEach
    void setup() {
        underTest = new JWKGenerator();
    }

    @Test
    void testPublicEncryptionKey() {

        // When
        final JWKPair encryptionKeyPair = underTest.generateEncryptionKeyPair(2048);

        final JWK publicKey = encryptionKeyPair.getPublicKey();

        // Then
        assertThat(publicKey, CoreMatchers.is(notNullValue()));


        assertThat(publicKey.getKeyID(), CoreMatchers.is(notNullValue()));

        assertThat(publicKey.getKeyOperations(), hasSize(1));
        assertThat(publicKey.getKeyOperations(), contains(KeyOperation.WRAP_KEY));

        assertThat(publicKey.getX509CertChain(), hasSize(1));

        assertThat(publicKey.getKeyType(), CoreMatchers.is(KeyType.RSA));
        assertThat( publicKey.getAlgorithm(), CoreMatchers.is(JWEAlgorithm.RSA_OAEP_256));
    }

    @Test
    void testPrivateDecryptionKey() {

        // When
        final JWKPair encryptionKeyPair = underTest.generateEncryptionKeyPair(2048);

        final JWK privateKey = encryptionKeyPair.getPrivateKey();

        // Then
        assertThat(privateKey, CoreMatchers.is(CoreMatchers.notNullValue()));

        assertThat(privateKey.getKeyID(), CoreMatchers.is(notNullValue()));

        assertThat(privateKey.getKeyOperations(), hasSize(1));
        assertThat(privateKey.getKeyOperations(), contains(KeyOperation.UNWRAP_KEY));

        assertThat(privateKey.getX509CertChain(), CoreMatchers.is(nullValue()));

        assertThat(privateKey.getKeyType(), CoreMatchers.is(KeyType.RSA));
        assertThat(privateKey.getAlgorithm(), CoreMatchers.is(JWEAlgorithm.RSA_OAEP_256));

        final Map<String, Object> keyParams = privateKey.toRSAKey().toJSONObject();

        assertThat(keyParams.get("d"), CoreMatchers.is(CoreMatchers.notNullValue()));
        assertThat(keyParams.get("dp"), CoreMatchers.is(CoreMatchers.notNullValue()));
        assertThat(keyParams.get("dq"), CoreMatchers.is(CoreMatchers.notNullValue()));
        assertThat(keyParams.get("e"), CoreMatchers.is(CoreMatchers.notNullValue()));
        assertThat(keyParams.get("n"), CoreMatchers.is(CoreMatchers.notNullValue()));
        assertThat(keyParams.get("p"), CoreMatchers.is(CoreMatchers.notNullValue()));
        assertThat(keyParams.get("q"), CoreMatchers.is(CoreMatchers.notNullValue()));
        assertThat(keyParams.get("qi"), CoreMatchers.is(CoreMatchers.notNullValue()));
    }

    @Test
    void testEncryptionAndDecryption() {

        // Given
        final JWKPair encryptionKeyPair = underTest.generateEncryptionKeyPair(4096);

        final JWK publicKey = encryptionKeyPair.getPublicKey();
        final JWK privateKey = encryptionKeyPair.getPrivateKey();

        final var data = "test string to encrypt";

        // When
        final JWECryptoService cryptoService = new JWECryptoService(null);

        final String encryptedData = cryptoService.encryptBytes(publicKey.toRSAKey(), data.getBytes(StandardCharsets.UTF_8));
        final byte[] decryptedData = cryptoService.decryptToBytes(privateKey.toRSAKey(), encryptedData);

        // Then
        assertThat(data, is(new String(decryptedData)));
    }

    @Test
    void testPublicEncryptionKeyValidationWithCorrectKeyLength() {

        // Given
        final JWKPair encryptionKeyPair = underTest.generateEncryptionKeyPair(4096);

        final JWK publicKey = encryptionKeyPair.getPublicKey();

        // Then
        assertDoesNotThrow(() -> JWKValidator.withoutX5CValidation()
                .withErrorLogLevel(LogLevel.ERROR)
                .build()
                .validate(publicKey.toRSAKey(), KeyOperation.WRAP_KEY));
    }

    @Test
    void testPublicSignatureVerificationKey() {

        // When
        final JWKPair signatureKeyPair = underTest.generateSignatureKeyPair(2048);

        final JWK publicKey = signatureKeyPair.getPublicKey();

        // Then
        assertThat(publicKey, CoreMatchers.is(CoreMatchers.notNullValue()));

        assertThat(publicKey.getKeyID(), CoreMatchers.is(notNullValue()));

        assertThat(publicKey.getKeyOperations(), hasSize(1));
        assertThat(publicKey.getKeyOperations(), contains(KeyOperation.VERIFY));

        assertThat(publicKey.getX509CertChain(), hasSize(1));

        assertThat(publicKey.getKeyType(), CoreMatchers.is(KeyType.RSA));
        assertThat( publicKey.getAlgorithm(), CoreMatchers.is(JWSAlgorithm.PS512));

    }

    @Test
    void testPublicSignatureKeyValidation() {

        // Given
        final JWKPair signatureKeyPair = underTest.generateSignatureKeyPair(4096);

        final JWK publicKey = signatureKeyPair.getPublicKey();

        // Then
        assertDoesNotThrow(() -> JWKValidator.withoutX5CValidation()
                .withErrorLogLevel(LogLevel.ERROR)
                .build()
                .validate(publicKey.toRSAKey(), KeyOperation.VERIFY));
    }

    @Test
    void testPrivateSigningKey() {

        // When
        final JWKPair signatureKeyPair = underTest.generateSignatureKeyPair(2048);

        final JWK privateKey =  signatureKeyPair.getPrivateKey();

        // Then
        assertThat(privateKey, CoreMatchers.is(notNullValue()));


        assertThat(privateKey.getKeyID(), CoreMatchers.is(notNullValue()));

        assertThat(privateKey.getKeyOperations(), hasSize(1));
        assertThat(privateKey.getKeyOperations(), contains(KeyOperation.SIGN));

        assertThat(privateKey.getX509CertChain(), CoreMatchers.is(nullValue()));

        assertThat(privateKey.getKeyType(), CoreMatchers.is(KeyType.RSA));
        assertThat(privateKey.getAlgorithm(), CoreMatchers.is(JWSAlgorithm.PS512));

        final Map<String, Object> keyParams = privateKey.toRSAKey().toJSONObject();

        assertThat(keyParams.get("d"), CoreMatchers.is(CoreMatchers.notNullValue()));
        assertThat(keyParams.get("dp"), CoreMatchers.is(CoreMatchers.notNullValue()));
        assertThat(keyParams.get("dq"), CoreMatchers.is(CoreMatchers.notNullValue()));
        assertThat(keyParams.get("e"), CoreMatchers.is(CoreMatchers.notNullValue()));
        assertThat(keyParams.get("n"), CoreMatchers.is(CoreMatchers.notNullValue()));
        assertThat(keyParams.get("p"), CoreMatchers.is(CoreMatchers.notNullValue()));
        assertThat(keyParams.get("q"), CoreMatchers.is(CoreMatchers.notNullValue()));
        assertThat(keyParams.get("qi"), CoreMatchers.is(CoreMatchers.notNullValue()));

    }
}